package article

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"testing"
)

type Key struct {
	X int
	Y string
}

func TestComparison(t *testing.T) {
	testcases := []struct {
		Name    string
		Error   bool // true if the comparison for equality should fail
		Pointer bool // true if we should use pointer semantics
		A       interface{}
		B       interface{}
	}{
		{
			Name: "Simple",
			A:    Key{X: 100, Y: "hello"},
			B:    Key{X: 100, Y: "hello"},
		},
		{
			Name:  "Different Content 1",
			Error: true,
			A:     Key{X: 1, Y: "hello"},
			B:     Key{X: 100, Y: "hello"},
		},
		{
			Name:  "Different Content 2",
			Error: true,
			A:     Key{X: 100, Y: "hello"},
			B:     Key{X: 100, Y: "world"},
		},
		{
			Name:    "Different Pointers",
			Error:   true,
			Pointer: true,
			A:       &Key{X: 100, Y: "hello"},
			B:       &Key{X: 100, Y: "hello"},
		},
	}

	for _, tc := range testcases {
		tc := tc
		t.Run(tc.Name, func(t *testing.T) {
			if tc.Error {
				if tc.A == tc.B {
					t.Errorf(`tc.A and tc.B should not match`)
					return
				}
			} else {
				// tc.A == tc.B -> OK
				if tc.A != tc.B {
					t.Errorf(`tc.A and tc.B should match`)
					return
				}
			}
		})
	}
}

type EmptyKey struct{}

func BenchmarkKey(b *testing.B) {
	m := make(map[interface{}]interface{})
	const strKey = "abracadabra Alakazm Hocus Pocus Open Sesame Sim Sala Bim"
	m[strKey] = 1
	m[Key{X: 100, Y: "hello"}] = 1
	m[EmptyKey{}] = 1

	check := func(b *testing.B, m map[interface{}]interface{}, key interface{}) {
		b.Helper()

		v, ok := m[key]
		if !ok {
			panic(fmt.Sprintf("Key %#v not found", key))
		}
		fmt.Fprintf(ioutil.Discard, "%s", v)
	}

	b.Run("String (const)", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			check(b, m, strKey)
		}
	})
	b.Run("String (non-const)", func(b *testing.B) {
		var buf bytes.Buffer
		buf.WriteString("abracadabra Alakazm Hocus Pocus Open Sesame Sim Sala Bim")
		for i := 0; i < b.N; i++ {
			check(b, m, buf.String())
		}
	})
	b.Run("Struct", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			check(b, m, Key{X: 100, Y: "hello"})
		}
	})
	b.Run("Empty Struct", func(b *testing.B) {
		for i := 0; i < b.N; i++ {
			check(b, m, EmptyKey{})
		}
	})
}
